#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
[[[
    One method, called magic sets,is a general algorithm for rewriting logical rules
so that they may be implemented bottom-UP (= forward chaining) in a way that
is that by working bottom-up, we can take advantage of efficient methods for doing
massive joins.
]]] -- Magic Sets and Other Strange Ways to Implement Logic Programs, F. Bancilhon,
D. Maier, Y. Sagiv and J. Ullman, Proc. 5th ACM SIGMOD-SIGACT Symposium on
Principles of Database Systems, 1986.

" [..] proposed transformation is to define additional predicates that compute the values that 
are passed from one predicate to another in the original rules, according to the sip strategy chosen for each
rule. Each of the original rules is then modified so that it fires only when values for these
additional predicates are available. These auxiliary predicates are called "magic predicates" and
the sets of values that they compute are called "magic sets". The intention is that the bottom-up
evaluation of the modified set of rules simulate the sip we have chosen for each adorned rule,
thus restricting the search space."

" we are interested in binding propagation and how it can be used to improve the
 efficiency of evaluation in the presence of recursion

"""
import unittest, os, time, itertools, copy
from FuXi.Rete.RuleStore import SetupRuleStore, N3RuleStore, N3Builtin, LOG
from FuXi.Rete.AlphaNode import ReteToken
from FuXi.Rete.BetaNode import project
from FuXi.Horn.HornRules import Clause, Ruleset, Rule
from FuXi.DLP.ConditionalAxioms import AdditionalRules
from FuXi.Horn.PositiveConditions import *
from FuXi.Rete.Proof import *
from FuXi.Syntax.InfixOWL import OWL_NS
from cStringIO import StringIO
from rdflib.Collection import Collection
from rdflib.Graph import Graph, ReadOnlyGraphAggregate
from rdflib import URIRef, RDF, RDFS, Namespace, Variable, Literal, URIRef
from rdflib.sparql.Algebra import RenderSPARQLAlgebra
from rdflib.sparql.parser import parse
from rdflib.util import first
from FuXi.Rete.SidewaysInformationPassing import *

EX_ULMAN = Namespace('http://doi.acm.org/10.1145/6012.15399#')
LOG_NS   = Namespace("http://www.w3.org/2000/10/swap/log#")
MAGIC = Namespace('http://doi.acm.org/10.1145/28659.28689#')

DDL_STRICTNESS_LOOSE    = 0
DDL_STRICTNESS_HARSH    = 1
DDL_STRICTNESS_FALLBACK_DERIVED = 2
DDL_STRICTNESS_FALLBACK_BASE = 3 
DDL_MUST_CHECK = [DDL_STRICTNESS_HARSH,
                  DDL_STRICTNESS_FALLBACK_DERIVED,
                  DDL_STRICTNESS_FALLBACK_BASE]
DDL_FALLBACK = [
    DDL_STRICTNESS_FALLBACK_DERIVED,
    DDL_STRICTNESS_FALLBACK_BASE
]

nameMap = {
  'loose'          : DDL_STRICTNESS_LOOSE,
  'defaultDerived' : DDL_STRICTNESS_FALLBACK_DERIVED,
  'defaultBase'    : DDL_STRICTNESS_FALLBACK_BASE,
  'harsh'          : DDL_STRICTNESS_HARSH,
}

def SetupDDLAndAdornProgram(factGraph,
                            rules,
                            GOALS,
                            derivedPreds=None,
                            strictCheck=DDL_STRICTNESS_FALLBACK_DERIVED,
                            defaultPredicates=None,
                            ignoreUnboundDPreds=False):
    if not defaultPredicates:
        defaultPredicates = [],[]
    if not derivedPreds:
        _derivedPreds=DerivedPredicateIterator(factGraph,
                                               rules,
                                               strict=strictCheck,
                                               defaultPredicates=defaultPredicates)
        if not isinstance(derivedPreds,(set,list)):
            derivedPreds=list(_derivedPreds)
        else:
            derivedPreds.extend(_derivedPreds)
    adornedProgram = AdornProgram(factGraph,rules,GOALS,derivedPreds,ignoreUnboundDPreds)
    if factGraph is not None:
        factGraph.adornedProgram = adornedProgram    
    return adornedProgram

def MagicSetTransformation(factGraph,
                           rules,
                           GOALS,
                           derivedPreds=None,
                           strictCheck=DDL_STRICTNESS_FALLBACK_DERIVED,
                           noMagic=None,
                           defaultPredicates=None):
    """
    Takes a goal and a ruleset and returns an iterator
    over the rulest that corresponds to the magic set
    transformation:
    """
    noMagic = noMagic and noMagic or []
    magicPredicates=set()
    replacement={}
    adornedProgram = SetupDDLAndAdornProgram(
                                   factGraph,
                                   rules,
                                   GOALS,
                                   derivedPreds=derivedPreds,
                                   strictCheck=strictCheck,
                                   defaultPredicates=defaultPredicates)
    newRules=[]
    for rule in adornedProgram: 
        if rule.isSecondOrder():
            import warnings
            warnings.warn(
            "Second order rule no supported by GMS: %s"%rule,
                    RuntimeWarning)
            
        magicPositions={}
        #Generate magic rules
        for idx,pred in enumerate(iterCondition(rule.formula.body)):
            magicBody=[]
            if isinstance(pred,AdornedUniTerm):# and pred not in magicPredicates:
                # For each rule r in Pad, and for each occurrence of an adorned 
                # predicate p a in its body, we generate a magic rule defining magic_p a
                prevPreds=[item for _idx,item in enumerate(rule.formula.body)
                                            if _idx < idx]             
                if 'b' not in pred.adornment:
                    import warnings
                    warnings.warn(
"adorned predicate w/out any bound arguments(%s in %s) !"%(pred,rule.formula),
                            RuntimeWarning)
                if GetOp(pred) not in noMagic:
                    magicPred=pred.makeMagicPred()
                    magicPositions[idx]=(magicPred,pred)
                    inArcs=[(N,x) for (N,x) in IncomingSIPArcs(rule.sip,getOccurrenceId(pred))
                                        if not set(x).difference(GetArgs(pred))]
                    if len(inArcs) > 1:
    #                    If there are several arcs entering qi, we define the magic rule defining magic_qi
    #                    in two steps. First, for each arc Nj --> qi with label cj , we define a rule with head label_qi_j(cj ). The body
    #                    of the rule is the same as the body of the magic rule in the case where there is a single arc
    #                    entering qi (described above). Then the magic rule is defined as follows. The head is
    #                    magic_q(0). The body contains label_qi_j(cj) for all j (that is, for all arcs entering qi ).
    
                        #We combine all incoming arcs into a single list of (body) conditions for the magic set
                        PrettyPrintRule(rule)
                        SIPRepresentation(rule.sip)
                        print pred, magicPred
                        _body=[]
                        additionalRules=[]
                        for idxSip,(N,x) in enumerate(inArcs):
                            newPred=pred.clone()
                            SetOp(newPred,URIRef('%s_label_%s'%(newPred.op,idxSip)))
                            ruleBody=And(buildMagicBody(
                                    N,
                                    prevPreds,
                                    rule.formula.head,
                                    derivedPreds))
                            additionalRules.append(Rule(Clause(ruleBody,newPred)))
                            _body.extend(newPred)                        
    #                        _body.extend(ruleBody)
                        additionalRules.append(Rule(Clause(And(_body),magicPred)))
                        newRules.extend(additionalRules)
                        for i in additionalRules:
                            print i
                        raise NotImplementedError()
                    else:
                        for idxSip,(N,x) in enumerate(inArcs):
                            ruleBody=And(buildMagicBody(
                                    N,
                                    prevPreds,
                                    rule.formula.head,
                                    derivedPreds,
                                    noMagic))
                            newRule=Rule(Clause(ruleBody,magicPred))
                            newRules.append(newRule)
                    magicPredicates.add(magicPred)
        #Modify rules
        #we modify the original rule by inserting
        #occurrences of the magic predicates corresponding
        #to the derived predicates of the body and to the head
        #If there are no bound arguments in the head, we don't modify the rule
        idxIncrement=0
        newRule=copy.deepcopy(rule)
        for idx,(magicPred,origPred) in magicPositions.items():
            newRule.formula.body.formulae.insert(idx+idxIncrement,magicPred)
            idxIncrement+=1
        if 'b' in rule.formula.head.adornment and GetOp(rule.formula.head) not in noMagic:
            headMagicPred=rule.formula.head.makeMagicPred()
            if isinstance(newRule.formula.body,Uniterm):
                newRule.formula.body = And([headMagicPred,newRule.formula.body])
            else:
                newRule.formula.body.formulae.insert(0,headMagicPred)
        newRules.append(newRule)

    if not newRules:
        newRules.extend(AdditionalRules(factGraph))
    for rule in newRules:
        if rule.formula.body:
            yield rule
        
def NormalizeGoals(goals):
    if isinstance(goals,(list,set)):
        for goal in goals:
            yield goal,{}
    elif isinstance(goals,tuple):
        yield sparqlQuery,{}
    else:
        query=RenderSPARQLAlgebra(parse(goals))
        for pattern in query.patterns:
            yield pattern[:3],query.prolog.prefixBindings
    
class AdornedRule(Rule):
    """Rule with 'bf' adornment and is comparable"""
    def __init__(self, clause, declare=None,nsMapping=None):
        decl=set()
        self.ruleStr=''
        for pred in itertools.chain(iterCondition(clause.head),
                                    iterCondition(clause.body)):
            decl.update([term for term in GetArgs(pred) if isinstance(term,Variable)])
            if isinstance(pred,AdornedUniTerm):
                self.ruleStr+=''.join(pred.adornment)
            self.ruleStr+=''.join(pred.toRDFTuple())
        super(AdornedRule, self).__init__(clause,decl,nsMapping)        

    def isRecursive(self):
        def termHash(term):
            return GetOp(term),\
                   reduce(lambda x,y:x+y,term.adornment)
        headHash = termHash(self.formula.head)
        def recursiveLiteral(term):
            return isinstance(term,AdornedUniTerm) and termHash(term) == headHash        
        if first(itertools.ifilter(recursiveLiteral,iterCondition(self.formula.body))):
            return True
        else:
            return False

    def __hash__(self):
        return hash(self.ruleStr)
        
    def __eq__(self,other):
        return hash(self) == hash(other)   

def NormalizeUniterm(term):
    if isinstance(term,Uniterm):
        return term
    elif isinstance(term,N3Builtin):
        return Uniterm(term.uri,term.argument,term.naf) 
    
def AdornRule(derivedPreds,clause,newHead,ignoreUnboundDPreds = False):
    """
    Adorns a horn clause using the given new head and list of
    derived predicates
    """
    assert len(list(iterCondition(clause.head)))==1
    adornedHead=AdornedUniTerm(clause.head,
                               newHead.adornment)
    sip=BuildNaturalSIP(clause,
                        derivedPreds,
                        adornedHead,
                        ignoreUnboundDPreds=ignoreUnboundDPreds)
    bodyPredReplace={}
    def adornment(arg,headArc,x):
        if headArc:
            #Sip arc from head
            #don't mark bound if query has no bound/distinguished terms
            return (arg in x and 
                    arg in adornedHead.getDistinguishedVariables(True)) \
                        and 'b' or 'f'
        else:
            return arg in x and 'b' or 'f'
    for literal in iterCondition(sip.sipOrder):
        op   = GetOp(literal)
        args = GetArgs(literal)
        if op in derivedPreds:
            for N,x in IncomingSIPArcs(sip,getOccurrenceId(literal)): 
                headArc = len(N)==1 and N[0] == GetOp(newHead)                
                if not set(x).difference(args):
                    # A binding
                    # for q is useful, however, only if it is a binding for an argument of q.
                    bodyPredReplace[literal]=AdornedUniTerm(NormalizeUniterm(literal),
                            [ adornment(arg,headArc,x) for arg in args],literal.naf)
#                For a predicate occurrence with no incoming
#                arc, the adornment contains only f. For our purposes here, 
#                we do not distinguish between a predicate with such an 
#                adornment and an unadorned predicate (we do in order to support open queries)
            if literal not in bodyPredReplace and ignoreUnboundDPreds:
                bodyPredReplace[literal]=AdornedUniTerm(NormalizeUniterm(literal),
                        [ 'f' for arg in GetArgs(literal)],literal.naf)
    rule=AdornedRule(Clause(And([bodyPredReplace.get(p,p) 
                                for p in iterCondition(sip.sipOrder)]),
                            adornedHead))
    rule.sip = sip
    return rule

def compareAdornedPredToRuleHead(aPred,head):
    """
    If p_a is an unmarked adorned predicate, then for each rule that has p in its head, ..
    """
    headPredicateTerm = GetOp(head)
    aPredTerm         = GetOp(aPred)
    assert isinstance(head,Uniterm)
    if head.getArity() == aPred.getArity():
        return headPredicateTerm == aPredTerm or isinstance(headPredicateTerm,
                                                            Variable)
    return False

def AdornProgram(factGraph,rs,goals,derivedPreds=None,ignoreUnboundDPreds=False):
    """
    The process starts from the given query. The query determines bindings for q, and we replace
    q by an adorned version, in which precisely the positions bound in the query are designated as
    bound, say q e . In general, we have a collection of adorned predicates, and as each one is processed,
    we will mark it, so that it will not be processed again. If p a is an unmarked adorned
    predicate, then for each rule that has p in its head, we generate an adorned version for the rule
    and add it to Pad; then p is marked as processed.    
    
    The adorned version of a rule contains additional
    adorned predicates, and these are added to the collection, unless they already appear
    there. The process terminates when no unmarked adorned predicates are left.
        
    >>> ruleStore,ruleGraph=SetupRuleStore(StringIO(PROGRAM2))
    >>> ruleStore._finalize()
    >>> fg=Graph().parse(StringIO(PROGRAM2),format='n3')
    >>> rs,query=AdornProgram(fg,ruleGraph,NON_LINEAR_MS_QUERY)
    >>> for rule in rs: print rule
    Forall ?Y ?X ( ex:sg(?X ?Y) :- ex:flat(?X ?Y) )
    Forall ?Y ?Z4 ?X ?Z1 ?Z2 ?Z3 ( ex:sg(?X ?Y) :- And( ex:up(?X ?Z1) ex:sg(?Z1 ?Z2) ex:flat(?Z2 ?Z3) ex:sg(?Z3 ?Z4) ex:down(?Z4 ?Y) ) )
    >>> print query
    (rdflib.URIRef('http://doi.acm.org/10.1145/6012.15399#john'), rdflib.URIRef('http://doi.acm.org/10.1145/6012.15399#sg'), ?X)
    """
    from FuXi.DLP import LloydToporTransformation
#    rs=rs is None and Ruleset(n3Rules=ruleGraph.store.rules,
#               nsMapping=ruleGraph.store.nsMgr) or rs
    
    adornedPredicateCollection=set()
    for goal,nsBindings in NormalizeGoals(goals):
        adornedPredicateCollection.add(AdornLiteral(goal,nsBindings))
    if not derivedPreds:
        derivedPreds=list(DerivedPredicateIterator(factGraph,rs))
    def unprocessedPreds(aPredCol):
        rt=[]
        for p in aPredCol:
            if not  p.marked:
                rt.append(p)
        return rt
    toDo=unprocessedPreds(adornedPredicateCollection)
    adornedProgram=set()
    while toDo:
        for term in toDo:
            #check if there is a rule with term as its head
            for rule in rs:
                for clause in LloydToporTransformation(rule.formula):
                    head=isinstance(clause.head,Exists) and clause.head.formula or clause.head
                    headPredicate = GetOp(head)
                    if compareAdornedPredToRuleHead(term,head):
                        #for each rule that has p in its head, we generate an adorned version for the rule
                        adornedRule=AdornRule(derivedPreds,
                                              clause,
                                              term,
                                              ignoreUnboundDPreds=ignoreUnboundDPreds)                        
                        adornedProgram.add(adornedRule)
                        #The adorned version of a rule contains additional adorned
                        #predicates, and these are added
                        for pred in iterCondition(adornedRule.formula.body):
                            if isinstance(pred,N3Builtin):
                                aPred = pred
                            else:
                                aPred = not isinstance(pred,AdornedUniTerm) and \
                                            AdornLiteral(pred.toRDFTuple(),
                                                         nsBindings,
                                                         pred.naf) or pred 
                            if GetOp(pred) in derivedPreds and aPred not in adornedPredicateCollection:
                                adornedPredicateCollection.add(aPred)
            term.marked=True
        toDo=unprocessedPreds(adornedPredicateCollection)
    return adornedProgram

class AdornedUniTerm(Uniterm):
    def __init__(self,uterm,adornment=None,naf = False):
        self.marked = False
        self.adornment=adornment
        self.nsMgr=uterm.nsMgr
        newArgs=copy.deepcopy(uterm.arg)
        super(AdornedUniTerm, self).__init__(uterm.op,newArgs,naf=naf)
        self.isMagic=False

    def clone(self):
        return AdornedUniTerm(self,self.adornment,self.naf)
                
    def makeMagicPred(self):
        """
        Make a (cloned) magic predicate
        
        The arity of the new predicate is the number of occurrences of b in the 
        adornment a, and its arguments correspond to the bound arguments of p a
        """
        newAdornedPred=AdornedUniTerm(self,self.adornment,self.naf)
        if self.op == RDF.type:
            newAdornedPred.arg[-1] = URIRef(self.arg[-1]+'_magic')
        elif len([i for i in self.adornment if i =='b'])==1:
            #adorned predicate occurrence with one out of two arguments bound
            #converted into a magic predicate: It becomes a unary predicate 
            #(an rdf:type assertion)
            newAdornedPred.arg[-1] = URIRef(self.op+'_magic')
            newAdornedPred.arg[0] = [self.arg[idx] 
                                        for idx,i in enumerate(self.adornment) 
                                                if i =='b'][0]            
            newAdornedPred.op = RDF.type
        else:
            newAdornedPred.op=URIRef(self.op+'_magic')
        newAdornedPred.isMagic=True
        return newAdornedPred

    def __hash__(self):
        return self._hash ^ hash(reduce(lambda x,y:x+y,self.adornment))
        
    # def __eq__(self,other):
    #     return self.adornment==other.adornment and\
    #            self.op==other.op and\
    #            self.arg==other.arg
                
    def hasBindings(self):
        for idx,term in enumerate(GetArgs(self)):
            if self.adornment[idx]=='b':
                if not varsOnly or isinstance(term,Variable):
                    return True
        return False
                
    def getDistinguishedVariables(self,varsOnly=False):
        if self.op == RDF.type:
            for idx,term in enumerate(GetArgs(self)):
                if self.adornment[idx]=='b':
                    if not varsOnly or isinstance(term,Variable):
                        yield term
        else:
            for idx,term in enumerate(self.arg):
                try:
                    if self.adornment[idx]=='b':
                        if not varsOnly or isinstance(term,Variable):
                            yield term
                except IndexError: pass
                  
    def getBindings(self,uniterm):
        rt={}
        for idx,term in enumerate(self.arg):
            goalArg=self.arg[idx]
            candidateArg=uniterm.arg[idx]
            if self.adornment[idx]=='b' and isinstance(candidateArg,Variable):
                #binding
                rt[candidateArg]=goalArg
        return rt
        
    def toRDFTuple(self):
        if hasattr(self,'isMagic') and self.isMagic:
            return (self.arg[0],self.op,self.arg[-1])
        else:
            subject,_object = self.arg
            return (subject,self.op,_object)
                
    def __repr__(self):
        pred = self.normalizeTerm(self.op)
        negPrefix = self.naf and 'not ' or ''
        if self.op == RDF.type:
            adornSuffix = '_'+self.adornment[0]
        else:
            adornSuffix='_'+''.join(self.adornment)
        if self.isMagic:
            if self.op == RDF.type:
                return "%s%s(%s)"%(negPrefix,
                                   self.normalizeTerm(self.arg[-1]),
                                   self.normalizeTerm(self.arg[0]))
            else:
                return "%s%s(%s)"%(negPrefix,
                                   pred,
                                ' '.join([self.normalizeTerm(i) 
                                            for idx,i in enumerate(self.arg) 
                                                    if self.adornment[idx]=='b']))
        elif self.op == RDF.type:
            return "%s%s%s(%s)"%(negPrefix,
                                 self.normalizeTerm(self.arg[-1]),
                                 adornSuffix,
                                 self.normalizeTerm(self.arg[0]))
        else:
            return "%s%s%s(%s)"%(negPrefix,
                                 pred,
                                 adornSuffix,
                                 ' '.join([self.normalizeTerm(i) 
                                            for i in self.arg]))

def AdornLiteral(rdfTuple,newNss=None,naf = False):
    """
    An adornment for an n-ary predicate p is a string a of length n on the 
    alphabet {b, f}, where b stands for bound and f stands for free. We 
    assume a fixed order of the arguments of the predicate.
    
    Intuitively, an adorned occurrence of the predicate, p a, corresponds to a 
    computation of the predicate with some arguments bound to constants, and 
    the other arguments free, where the bound arguments are those that are
    so indicated by the adornment.    
    
    >>> EX=Namespace('http://doi.acm.org/10.1145/6012.15399#')
    >>> query=RenderSPARQLAlgebra(parse(NON_LINEAR_MS_QUERY))
    >>> literal=query.patterns[0][:3]
    >>> literal
    (rdflib.URIRef('http://doi.acm.org/10.1145/6012.15399#john'), rdflib.URIRef('http://doi.acm.org/10.1145/6012.15399#sg'), ?X)
    >>> aLit=AdornLiteral(literal,query.prolog.prefixBindings)
    >>> aLit
    mst:sg_bf(mst:john ?X)
    >>> aLit.adornment
    ['b', 'f']
    >>> aLit.getBindings(Uniterm(EX.sg,[Variable('X'),EX.jill]))
    {?X: rdflib.URIRef('http://doi.acm.org/10.1145/6012.15399#john')}
    """
    args=[rdfTuple[0],rdfTuple[-1]]
    newNss=newNss is None and {} or newNss
    uTerm = BuildUnitermFromTuple(rdfTuple,newNss)
    opArgs=rdfTuple[1] == RDF.type and [args[0]] or args
    def isFreeTerm(term):
        return isinstance(term,Variable)
    adornment=[ isFreeTerm(term) and 'f' or 'b' for idx,term in enumerate(opArgs) ]
    return AdornedUniTerm(uTerm,adornment,naf)  

def DerivedPredicateIterator(factsOrBasePreds,
                             ruleset,
                             strict=DDL_STRICTNESS_FALLBACK_DERIVED,
                             defaultPredicates=None):
    """
    >>> ruleStore,ruleGraph=SetupRuleStore()
    >>> g=ruleGraph.parse(StringIO(MAGIC_PROGRAM1),format='n3')
    >>> ruleStore._finalize()
    >>> ruleFacts=Graph().parse(StringIO(MAGIC_PROGRAM1),format='n3')
    >>> for lit in DerivedPredicateIterator(ruleFacts,ruleGraph): print lit
    ex:anc(?X ?Y)
    >>> ruleStore,ruleGraph=SetupRuleStore()
    >>> g=ruleGraph.parse(StringIO(PROGRAM2),format='n3')
    >>> ruleStore._finalize()
    >>> ruleFacts=Graph().parse(StringIO(PROGRAM2),format='n3')    
    >>> for lit in DerivedPredicateIterator(ruleFacts,ruleGraph): print lit
    ex:sg(?X ?Y)
    """
    if not defaultPredicates:
        defaultPredicates = [],[]
    defaultBasePreds,defaultDerivedPreds = defaultPredicates
    basePreds=[GetOp(buildUniTerm(fact)) for fact in factsOrBasePreds 
                        if fact[1] != LOG.implies]     
    processed={True:set(),False:set()}
    derivedPreds=set()
    uncertainPreds=set()
    ruleBodyPreds=set()
    ruleHeads=set()
    for rule in ruleset:
        if rule.formula.body:
            for idx,term in enumerate(itertools.chain(iterCondition(rule.formula.head),
                                      iterCondition(rule.formula.body))):
                #iterate over terms from head to end of body
                op = GetOp(term)
                if op not in processed[idx>0]: 
                    #not processed before
                    if idx > 0:
                        #body literal
                        ruleBodyPreds.add(op)
                    else:
                        #head literal
                        ruleHeads.add(op)
                    if strict in DDL_MUST_CHECK and \
                        not (op not in basePreds or idx > 0):
                        #checking DDL well formedness and
                        #op is a base predicate *and* a head literal (derived)
                        if strict in DDL_FALLBACK:
                            mark = strict == DDL_STRICTNESS_FALLBACK_DERIVED and \
                            'derived' or 'base'
                            if strict == DDL_STRICTNESS_FALLBACK_DERIVED and \
                                op not in defaultBasePreds:     
                                #a clashing predicate is marked as derived due
                                #to level of strictness                        
                                derivedPreds.add(op)
                            elif strict == DDL_STRICTNESS_FALLBACK_BASE and \
                                op not in defaultDerivedPreds:
                                #a clashing predicate is marked as base dur
                                #to level of strictness
                                defaultBasePreds.append(op)
                            import warnings
                            warnings.warn(
      "predicate symbol of %s is in both IDB and EDB. Marking as %s"%(term,mark))
                        else:
                            raise SyntaxError("%s is a member of a derived predicate and a base predicate!"%term)
                    if op in basePreds:
                        #base predicates are marked for later validation
                        uncertainPreds.add(op)
                    else:
                        if idx == 0 and not isinstance(op,Variable):
                            #head literal with proper predicate symbol
                            #identify as a derived predicate
                            derivedPreds.add(op)
                        elif not isinstance(op,Variable):
                            #body literal with proper predicate symbol
                            #mark for later validation
                            uncertainPreds.add(op)
                    processed[idx>0].add(op)
    for pred in uncertainPreds:
        #for each predicate marked as 'uncertain'
        #do further checking
        if (pred not in ruleBodyPreds and not isinstance(pred,Variable)) or\
           pred in ruleHeads:
            #pred is not in a body literal and is a proper predicate symbol
            #or it is a rule head -> mark as a derived predicate
            derivedPreds.add(pred)
    for pred in derivedPreds:
        if not pred in defaultBasePreds:
            yield pred
    
def iter_non_base_non_derived_preds(ruleset,intensional_db):
    rt=set()
    intensional_preds=set([p for p in intensional_db.predicates() 
                                    if p != LOG_NS.implies])
    for rule in ruleset:
        for uterm in rule.formula.head:
            if uterm.op in intensional_preds and uterm.op not in rt:
                rt.add(uterm.op)
                yield uterm.op, (fact 
                         for fact in intensional_db.triples((None,uterm.op,None)))

def NormalizeLPDb(ruleGraph,fact_db):
    """
    For performance reasons, it 1s good to decompose the database into a set of
    pure base predicates (which can then be stored using a standard DBMS)
    and a set of pure derived predicates Fortunately, such a decomposition 1s 
    always possible, because every database can be rewritten ...as 
    database containing only base and derived predicates.    
    
    >>> ruleStore,ruleGraph=SetupRuleStore()
    >>> g=ruleGraph.parse(StringIO(PARTITION_LP_DB_PREDICATES),format='n3')
    >>> ruleStore._finalize()    
    >>> len(ruleStore.rules)
    1
    >>> factGraph=Graph().parse(StringIO(PARTITION_LP_DB_PREDICATES),format='n3')
    >>> rs=Ruleset(n3Rules=ruleStore.rules,nsMapping=ruleStore.nsMgr)
    >>> for i in rs: print i
    Forall ?Y ?X ?Z ( ex:grandfather(?X ?Y) :- And( ex:father(?X ?Z) ex:parent(?X ?Y) ) )
    >>> len(factGraph)
    4
    >>> print [p for p,iter in iter_non_base_non_derived_preds(rs,factGraph)]
    [rdflib.URIRef('http://doi.acm.org/10.1145/16856.16859#grandfather')]
    """
    candidatePreds=False
    rs=Ruleset(n3Rules=ruleGraph.store.rules,
               nsMapping=ruleStore.nsMgr)
    toAdd=[]
    for pred,replFacts in iter_non_base_non_derived_preds(rs,fact_db):
        replPred=URIRef(pred+'_ext')
        for s,p,o in replFacts:
            fact_db.remove((s,p,o))
            toAdd.append((s,replPred,o))
        head=Uniterm(pred,pred.arg)
        body=Uniterm(replPred,pred.arg)
        newRule=Rule(Clause(body,head),
                     [term for term in pred.arg if isinstance(term,Variable)])
        rs.append(newRule)
    return rs

#class AdornProgramTest(unittest.TestCase):
#    def setUp(self):
#        self.ruleStore,self.ruleGraph=SetupRuleStore(StringIO(PROGRAM2))
#        self.ruleStore._finalize()
#        self.ruleStrings=[
#        'Forall ?Y ?X ( :sg_bf(?X ?Y) :- And( :sg_magic(?X) ex:flat(?X ?Y) ) )',
#        'Forall  ( :sg_magic(?Z1) :- And( :sg_magic(?X) ex:up(?X ?Z1) ) )',
#        'Forall ?Z4 ?Y ?X ?Z1 ?Z2 ?Z3 ( :sg_bf(?X ?Y) :- And( :sg_magic(?X) ex:up(?X ?Z1) :sg_magic(?Z1) :sg_bf(?Z1 ?Z2) ex:flat(?Z2 ?Z3) :sg_magic(?Z3) :sg_bf(?Z3 ?Z4) ex:down(?Z4 ?Y) ) )',
#        'Forall  ( :sg_magic(?Z3) :- And( :sg_magic(?X) ex:up(?X ?Z1) :sg_bf(?Z1 ?Z2) ex:flat(?Z2 ?Z3) ) )',
#        ]
#
#    def testAdorn(self):
#        fg=Graph().parse(StringIO(PROGRAM2),format='n3')
#        rules=Ruleset(n3Rules=self.ruleGraph.store.rules,
#                   nsMapping=self.ruleStore.nsMgr)
#        from pprint import pprint;pprint(self.ruleStrings)        
#        for rule in MagicSetTransformation(fg,
#                                           rules,
#                                           NON_LINEAR_MS_QUERY,
#                                           [MAGIC.sg]):
#            self.failUnless(repr(rule) in self.ruleStrings, repr(rule))

class AdornProgramTest(unittest.TestCase):
    def setUp(self):
        self.ruleStore,self.ruleGraph=SetupRuleStore(StringIO(PROGRAM2))
        self.ruleStore._finalize()
        self.ruleStrings=[
        'Forall ?Y ?X ( :sg_bf(?X ?Y) :- And( :sg_magic(?X) ex:flat(?X ?Y) ) )',
        'Forall  ( :sg_magic(?Z1) :- And( :sg_magic(?X) ex:up(?X ?Z1) ) )',
        'Forall ?Z4 ?Y ?X ?Z1 ?Z2 ?Z3 ( :sg_bf(?X ?Y) :- And( :sg_magic(?X) ex:up(?X ?Z1) :sg_magic(?Z1) :sg_bf(?Z1 ?Z2) ex:flat(?Z2 ?Z3) :sg_magic(?Z3) :sg_bf(?Z3 ?Z4) ex:down(?Z4 ?Y) ) )',
        'Forall  ( :sg_magic(?Z3) :- And( :sg_magic(?X) ex:up(?X ?Z1) :sg_bf(?Z1 ?Z2) ex:flat(?Z2 ?Z3) ) )',
        ]

    def testAdorn(self):
        fg=Graph().parse(StringIO(PROGRAM2),format='n3')
        rules=Ruleset(n3Rules=self.ruleGraph.store.rules,
                   nsMapping=self.ruleStore.nsMgr)
        from pprint import pprint;pprint(self.ruleStrings)        
        for rule in MagicSetTransformation(fg,
                                           rules,
                                           NON_LINEAR_MS_QUERY,
                                           [MAGIC.sg]):
            self.failUnless(repr(rule) in self.ruleStrings, repr(rule))
        
def buildMagicBody(N,prevPredicates,adornedHead,derivedPreds,noMagic=[]):
    unboundHead='b' in adornedHead.adornment
    if unboundHead:
        body=[adornedHead.makeMagicPred()]
    else:
        #If there are no bound argument positions to pass magic values with,
        #we propagate values in the full relation
        body=[]
    for prevAPred in prevPredicates:
        op = GetOp(prevAPred)
        if op in N or isinstance(op,Variable):
            #If qj, j<i, is in N, we add qj to the body of the magic rule
            #Note, if the atom has a variable for the predicate, treat it as a base
            #predicate occurrence
            body.append(prevAPred)
        if op in derivedPreds and isinstance(prevAPred,AdornedUniTerm) and prevAPred.adornment.count('b')>0:
            #If qj is a derived predicate and its adornment contains at least 
            #one b, we also add the corresponding magic predicate to the body
            if op in noMagic:
                body.append(prevAPred)
            else:
                body.append(prevAPred.makeMagicPred())
    return body

def PrettyPrintRule(rule):
    if isinstance(rule.formula.body,And):
        print rule.formula.head
        print "    :- %s"%rule.formula.body.formulae[0]
        for idx,literal in enumerate(rule.formula.body.formulae[1:]):
            print "       %s%s"%(literal,
                                 literal == rule.formula.body.formulae[-1] and '' or ', ')
    else:
        print rule.formula

OWL_PROPERTIES_QUERY=\
"""
SELECT ?prop
WHERE {
    ?prop a ?propType 
      FILTER( 
        ?propType = owl:ObjectProperty || 
        ?propType = owl:TransitiveProperty ||
        ?propType = owl:SymmetricProperty ||
        ?propType = owl:InverseFunctionalProperty ||
        ?propType = owl:DatatypeProperty )  
}"""

EXCLUDED_DERIVED_PREDS=\
[
    
]


def IdentifyDerivedPredicates(ddlMetaGraph,tBox,ruleset=None):
    """
    See: http://code.google.com/p/fuxi/wiki/DataDescriptionLanguage#
    """
    dPreds = set()
    basePreds = set()
    DDL = Namespace('http://code.google.com/p/fuxi/wiki/DataDescriptionLanguage#')
    
    if ruleset:
        for rule in ruleset:
            dPreds.add(GetOp(rule.formula.head))
    
    for derivedClassList in ddlMetaGraph.subjects(predicate=RDF.type,
                                          object=DDL.DerivedClassList):
        dPreds.update(Collection(ddlMetaGraph,derivedClassList))
    for derivedClassList in ddlMetaGraph.subjects(predicate=RDF.type,
                                          object=DDL.DerivedPropertyList):
        dPreds.update(Collection(ddlMetaGraph,derivedClassList))        
    derivedPropPrefixes = []
    basePropPrefixes    = []
    for derivedPropPrefixList in ddlMetaGraph.subjects(predicate=RDF.type,
                                               object=DDL.DerivedPropertyPrefix):
        derivedPropPrefixes.extend(Collection(ddlMetaGraph,derivedPropPrefixList))
    for basePropPrefixList in ddlMetaGraph.subjects(predicate=RDF.type,
                                               object=DDL.BasePropertyPrefix):
        basePropPrefixes.extend(Collection(ddlMetaGraph,basePropPrefixList))
        
    for prop in tBox.query(OWL_PROPERTIES_QUERY):
        if first(itertools.ifilter(lambda prefix:prop.startswith(prefix),
                                   derivedPropPrefixes)) and \
                       (prop,RDF.type,OWL_NS.AnnotationProperty) not in tBox: 
            dPreds.add(prop)
        if first(itertools.ifilter(lambda prefix:prop.startswith(prefix),
                                   basePropPrefixes)) and \
                       (prop,RDF.type,OWL_NS.AnnotationProperty) not in tBox and \
                       prop not in dPreds: 
            basePreds.add(prop)
            
    derivedClassPrefixes = []
    for derivedClsPrefixList in ddlMetaGraph.subjects(predicate=RDF.type,
                                              object=DDL.DerivedClassPrefix):
        derivedClassPrefixes.extend(Collection(ddlMetaGraph,
                                               derivedClsPrefixList))
    baseClassPrefixes = []
    for baseClsPrefixList in ddlMetaGraph.subjects(predicate=RDF.type,
                                             object=DDL.BaseClassPrefix):
       baseClassPrefixes.extend(Collection(ddlMetaGraph,
                                           baseClsPrefixList))
    for cls in tBox.subjects(predicate=RDF.type,
                             object=OWL_NS.Class):
        if first(itertools.ifilter(lambda prefix:cls.startswith(prefix),
                                   baseClassPrefixes)):
            if cls not in dPreds: 
                basePreds.add(cls)
        if first(itertools.ifilter(lambda prefix:cls.startswith(prefix),
                                   derivedClassPrefixes)):
            if cls not in basePreds: 
                dPreds.add(cls)
            
    nsBindings = dict([(prefix,nsUri) 
                       for prefix, nsUri in tBox.namespaces()
                        if prefix])            
    for queryNode in ddlMetaGraph.subjects(predicate=RDF.type,
                                   object=DDL.DerivedClassQuery):
        query = first(ddlMetaGraph.objects(queryNode,RDF.value))
        for cls in tBox.query(query,
                              initNs=nsBindings):
            dPreds.add(cls)
            
    for baseClsList in ddlMetaGraph.subjects(predicate=RDF.type,
                                             object=DDL.BaseClassList):
        basePreds.update(Collection(ddlMetaGraph,baseClsList))
                                             
    dPreds.difference_update(basePreds)
    return dPreds
def test():
    unittest.main()    
    # import doctest
    # doctest.testmod()

if __name__ == '__main__':
    test()


import doctest
from rdflib.Graph import Graph
from rdflib import RDF, URIRef, Namespace, Literal, BNode

def IdentifyHybridPredicates(graph,derivedPredicates):
    """
    Takes an RDF graph and a list of derived predicates and return
    those predicates that are both EDB (extensional) and IDB (intensional) predicates.
    i.e., derived predicates that appear in the graph
    
    >>> g=Graph()
    >>> EX= Namespace('http://example.com/')
    >>> g.add((BNode(),EX.predicate1,Literal(1)))
    >>> g.add((BNode(),RDF.type,EX.Class1))
    >>> g.add((BNode(),RDF.type,EX.Class2))
    >>> rt=IdentifyHybridPredicates(g,[EX.predicate1,EX.Class1,EX.Class3])
    >>> sorted(rt)
    [rdflib.URIRef('http://example.com/Class1'), rdflib.URIRef('http://example.com/predicate1')]
    """
    derivedPredicates = derivedPredicates if isinstance(derivedPredicates,
                                                        set) else \
                        set(derivedPredicates)
    return derivedPredicates.intersection(                    
                    [ o if p == RDF.type else p 
                        for s,p,o in graph ])    
    
if __name__ == '__main__':
    import doctest
    doctest.testmod()


#!/usr/local/bin/python
# -*- coding: utf-8 -*-
"""
Stratified Negation Semantics for DLP using SPARQL to handle the negation
"""
from pprint import pprint
from rdflib.Graph import Graph
from rdflib.util import first
from rdflib import RDF, RDFS, Namespace, Variable, Literal, URIRef, BNode
from rdflib.syntax.NamespaceManager import NamespaceManager
from FuXi.Rete.RuleStore import N3RuleStore,SetupRuleStore
from FuXi.Horn.PositiveConditions import And
from FuXi.Rete import ReteNetwork
from FuXi.Rete.RuleStore import N3RuleStore
from FuXi.Rete.Util import generateTokenSet
from FuXi.Syntax.InfixOWL import *
from FuXi.DLP import SKOLEMIZED_CLASS_NS, MapDLPtoNetwork
from DLNormalization import NormalFormReduction
import sys, unittest, copy, itertools

EX_NS = Namespace('http://example.com/')
EX    = ClassNamespaceFactory(EX_NS)

def GetVars(atom):
    from FuXi.Rete.SidewaysInformationPassing import GetArgs
    return [term for term in GetArgs(atom) if isinstance(term,Variable)]

def CalculateStratifiedModel(network,ontGraph,derivedPreds,edb=None):
    posRules,ignored=MapDLPtoNetwork(network,
                               ontGraph,
                               constructNetwork=False,
                               derivedPreds=derivedPreds,
                               ignoreNegativeStratus=True)
    for rule in posRules:
        network.buildNetworkFromClause(rule)    
    network.feedFactsToAdd(generateTokenSet(edb and edb or ontGraph))
    for i in ignored:
        #Evaluate the Graph pattern, and instanciate the head of the rule with 
        #the solutions returned
        sel,compiler=StratifiedSPARQL(i)
        query=compiler.compile(sel)
        i.stratifiedQuery=query
        vars = sel.projection
        for rt in (edb and edb or ontGraph).query(query):
            solutions={}
            if isinstance(rt,tuple):
                solutions.update(dict([(vars[idx],i) for idx,i in enumerate(rt)]))
            else:
                solutions[vars[0]]=rt
            i.solutions=solutions
            head=copy.deepcopy(i.formula.head)
            head.ground(solutions)
            fact=head.toRDFTuple()
            network.inferredFacts.add(fact)
            network.feedFactsToAdd(generateTokenSet([fact]))
            
    #Now we need to clear assertions that cross the individual, concept, relation divide
    toRemove=[]
    for s,p,o in network.inferredFacts.triples((None,
                                                RDF.type,
                                                None)):
        if s in (edb and edb or ontGraph).predicates() or\
           s in [_s for _s,_p,_o in 
                    (edb and edb or ontGraph).triples_choices(
                                        (None,
                                         RDF.type,
                                         [OWL_NS.Class,
                                          OWL_NS.Restriction]))]:
            network.inferredFacts.remove((s,p,o))
    return posRules,ignored
            
def createCopyPattern(toDo):
    """
    "Let φ : V → V be a variable-renaming function. Given a graph pattern P, a
    copy pattern φ(P) is an isomorphic copy of P whose variables have been renamed
    according to φ and satisfying that var(P) ∩ var(φ(P)) = ∅."    
    
    varExprs maps variable expressions to variables
    vars     maps variables to variables 
    
    """
    from telescope.sparql.helpers import v
    vars={}
    varExprs={}
    copyPatterns=[]
    for formula in toDo:
        for var in GetVars(formula):
            if not var in vars:
                newVar=Variable(BNode())
                varExprs[v[var]]=newVar
                vars[var]=newVar
        copyTriplePattern=copy.deepcopy(formula)
        copyTriplePattern.renameVariables(vars)
        copyPatterns.append(copyTriplePattern)
    return copyPatterns,vars,varExprs
            
def StratifiedSPARQL(rule,nsMapping={EX_NS: 'ex'}):
    """
    The SPARQL specification indicates that it is possible to test if a graph
    pattern does not match a dataset, via a combination of optional patterns and
    filter conditions (like negation as failure in logic programming)([9] Sec. 11.4.1).
    In this section we analyze in depth the scope and limitations of this approach.
    We will introduce a syntax for the “difference” of two graph patterns P1 and
    P2, denoted (P1 MINUS P2), with the intended informal meaning: “the set of
    mappings that match P1 and does not match P2”. 
    
    Uses telescope to construct the SPARQL MINUS BGP expressions for body 
    conditions with default negation formulae              
    """
    from FuXi.Rete.SidewaysInformationPassing import GetArgs, findFullSip, iterCondition
    #Find a sip order of the horn rule
    if isinstance(rule.formula.body,And):
        sipOrder=first(findFullSip(([rule.formula.head],None), rule.formula.body))
    else:
        sipOrder=[rule.formula.head]+[rule.formula.body]
    from telescope import optional, op
    from telescope.sparql.queryforms import Select
    from telescope.sparql.expressions import Expression
    from telescope.sparql.compiler import SelectCompiler
    from telescope.sparql.patterns import GroupGraphPattern 
    toDo=[]
    negativeVars=set()
    positiveLiterals = False
    for atom in sipOrder[1:]:
        if atom.naf:
            toDo.append(atom)
            negativeVars.update(GetVars(atom))
        else:
            positiveLiterals = True
    #The negative literas are moved to the back of the body conjunct
    #Intuitively, they should not be disconnected from the rest of rule
    #Due to the correlation between DL and guarded FOL
    [sipOrder.remove(toRemove) for toRemove in toDo]
    
    #posLiterals are all the positive literals leading up to the negated
    #literals (in left-to-right order)  There may be none, see below
    posLiterals=sipOrder[1:]    
    
    posVarIgnore = []
    if not positiveLiterals:
        #If there are no lead, positive literals (i.e. the LP is of the form:
        #   H :- not B1, not B2, ...
        #Then a 'phantom' triple pattern is needed as the left operand to the OPTIONAL
        #in order to properly implement P0 MINUS P where P0 is an empty pattern
        keyVar = GetVars(rule.formula.head)[0]
        newVar1=Variable(BNode())
        newVar2=Variable(BNode())
        posVarIgnore.extend([newVar1,newVar2])
        phantomLiteral=Uniterm(newVar1,[keyVar,newVar2])    
        posLiterals.insert(0,phantomLiteral)
            
    #The positive variables are collected
    positiveVars=set(reduce(lambda x,y:x+y,[GetVars(atom) for atom in posLiterals]))

    vars={}
    varExprs={}
    copyPatterns=[]
    print >> sys.stderr, "%s =: { %s MINUS %s} "%(rule.formula.head,
                                                  posLiterals,
                                                  toDo)
    def collapseMINUS(left,right):
        negVars=set()
        for pred in iterCondition(right):
            negVars.update([term for term in GetArgs(pred) 
                                    if isinstance(term,Variable)])
        innerCopyPatternNeeded = not negVars.difference(positiveVars)
        #A copy pattern is needed if the negative literals don't introduce new vars
        if innerCopyPatternNeeded:
            innerCopyPatterns,innerVars,innerVarExprs=createCopyPattern([right])
            #We use an arbitrary new variable as for the outer FILTER(!BOUND(..))
            outerFilterVariable=innerVars.values()[0]
            optionalPatterns=[right] +innerCopyPatterns
            negatedBGP=optional(*[formula.toRDFTuple() 
                                for formula in optionalPatterns])
            negatedBGP.filter(*[k==v for k,v in innerVarExprs.items()])         
            positiveVars.update([Variable(k.value[0:]) for k in innerVarExprs.keys()])
            positiveVars.update(innerVarExprs.values())   
        else:
            #We use an arbitrary, 'independent' variable for the outer FILTER(!BOUND(..))
            outerFilterVariable=negVars.difference(positiveVars).pop()
            optionalPatterns=[right]
            negatedBGP=optional(*[formula.toRDFTuple() 
                                for formula in optionalPatterns])
            positiveVars.update(negVars)
        left=left.where(*[negatedBGP])
        left=left.filter(~op.bound(outerFilterVariable))
        return left        
    topLevelQuery = Select(GetArgs(rule.formula.head)).where(
                             GroupGraphPattern.from_obj([
                                 formula.toRDFTuple() 
                                    for formula in posLiterals]))
    rt=reduce(collapseMINUS,[topLevelQuery]+toDo)
    return rt,SelectCompiler(nsMapping)

def ProperSipOrderWithNegation(body):
    """
    Ensures the list of literals has the negated literals
    at the end of the list
    """
    from FuXi.Rete.SidewaysInformationPassing import iterCondition
    #import pdb;pdb.set_trace()
    firstNegLiteral = None
    bodyIterator = list(body)
    for idx,literal in enumerate(bodyIterator):
        if literal.naf:
            firstNegLiteral = literal
            break
    if firstNegLiteral:
        #There is a first negative literal, are there subsequent positive literals?
        subsequentPosLits =  first(itertools.dropwhile(lambda i:i.naf,
                                                       bodyIterator[idx:]))
        if len(body) - idx > 1:
            #if this is not the last term in the body
            #then we succeed only if there are no subsequent positive literals
            return not subsequentPosLits 
        else:
            #this is the last term, so we are successful
            return True
    else:
        #There are no negative literals
        return True

class UniversalRestrictionTest(unittest.TestCase):
    def setUp(self):
        self.ontGraph = Graph()
        self.ontGraph.bind('ex', EX_NS)
        self.ontGraph.bind('owl', OWL_NS)
        Individual.factoryGraph = self.ontGraph
                
    def testNegatedDisjunctionTest(self):
        contains=Property(EX_NS.contains)
        omega = EX.Omega
        alpha = EX.Alpha
        innerDisjunct = omega | alpha
        foo = EX.foo
        testClass1 = foo & (contains|only|~innerDisjunct)
        testClass1.identifier = EX_NS.Bar
        
        self.assertEqual(repr(testClass1),
                "ex:foo that ( ex:contains only ( not ( ex:Omega or ex:Alpha ) ) )")
        NormalFormReduction(self.ontGraph)
        self.assertEqual(repr(testClass1),
                "ex:foo that ( not ( ex:contains some ( ex:Omega or ex:Alpha ) ) )")
        
        individual1 = BNode()
        individual2 = BNode()
        foo.extent = [individual1]
        contains.extent = [(individual1,individual2)]
        (EX.Baz).extent = [individual2]
        ruleStore,ruleGraph,network=SetupRuleStore(makeNetwork=True)
        posRules,ignored=CalculateStratifiedModel(network,self.ontGraph,[EX_NS.Bar])
        self.failUnless(not posRules,"There should be no rules in the 0 strata!")
        self.assertEqual(len(ignored),2,"There should be 2 'negative' rules")
        testClass1.graph = network.inferredFacts 
        self.failUnless(individual1 in testClass1.extent,
                        "%s should be in ex:Bar's extent"%individual1)        

    def testNominalPartition(self):
        partition = EnumeratedClass(EX_NS.part,
                                    members=[EX_NS.individual1,
                                             EX_NS.individual2,
                                             EX_NS.individual3])
        subPartition = EnumeratedClass(members=[EX_NS.individual1])
        partitionProp = Property(EX_NS.propFoo,
                                 range=partition.identifier)
        self.testClass = (EX.Bar) & (partitionProp|only|subPartition)
        self.testClass.identifier = EX_NS.Foo         
        self.assertEqual(repr(self.testClass),
                        "ex:Bar that ( ex:propFoo only { ex:individual1 } )")        
        self.assertEqual(repr(self.testClass.identifier),
                        "rdflib.URIRef('http://example.com/Foo')")        
        NormalFormReduction(self.ontGraph)
        self.assertEqual(repr(self.testClass),
        "ex:Bar that ( not ( ex:propFoo value ex:individual2 ) ) and ( not ( ex:propFoo value ex:individual3 ) )")
        ruleStore,ruleGraph,network=SetupRuleStore(makeNetwork=True)
        
        ex = BNode()
        (EX.Bar).extent = [ex]
        self.ontGraph.add((ex,EX_NS.propFoo,EX_NS.individual1))
        CalculateStratifiedModel(network,self.ontGraph,[EX_NS.Foo])
        self.failUnless((ex,RDF.type,EX_NS.Foo) in network.inferredFacts,
                        "Missing level 1 predicate (ex:Foo)")

class NegatedExistentialRestrictionTest(unittest.TestCase):
    def setUp(self):
        self.ontGraph = Graph()
        self.ontGraph.bind('ex', EX_NS)
        self.ontGraph.bind('owl', OWL_NS)
        Individual.factoryGraph = self.ontGraph
                
    def testInConjunct(self):
        contains=Property(EX_NS.contains)
        testCase2 = EX.Operation & ~ (contains|some|EX.IsolatedCABGConcomitantExclusion) &\
                                          (contains|some|EX.CoronaryArteryBypassGrafting)
        testCase2.identifier = EX_NS.IsolatedCABGOperation        
        NormalFormReduction(self.ontGraph)
        self.assertEqual(repr(testCase2),
                        "ex:Operation that ( ex:contains some ex:CoronaryArteryBypassGrafting ) and ( not ( ex:contains some ex:IsolatedCABGConcomitantExclusion ) )")
        ruleStore,ruleGraph,network=SetupRuleStore(makeNetwork=True)
        op=BNode()
        (EX.Operation).extent = [op]
        grafting=BNode()
        (EX.CoronaryArteryBypassGrafting).extent = [grafting]
        testCase2.graph.add((op,EX_NS.contains,grafting))        
        CalculateStratifiedModel(network,testCase2.graph,[EX_NS.Foo,EX_NS.IsolatedCABGOperation])
        testCase2.graph = network.inferredFacts 
        self.failUnless(op in testCase2.extent,
                        "%s should be in ex:IsolatedCABGOperation's extent"%op)        
        

    def testGeneralConceptInclusion(self):
#        Some Class 
#            ## Primitive Type  ##
#            SubClassOf: Class: ex:NoExclusion  . 
#            DisjointWith ( ex:contains some ex:IsolatedCABGConcomitantExclusion )
        contains=Property(EX_NS.contains)
        testClass = ~(contains|some|EX.Exclusion)
        testClass2 = EX.NoExclusion
        testClass2 += testClass
        NormalFormReduction(self.ontGraph)
        individual1 = BNode()
        individual2 = BNode()
        contains.extent = [(individual1,individual2)]
        ruleStore,ruleGraph,network=SetupRuleStore(makeNetwork=True)
        posRules,negRules=CalculateStratifiedModel(network,self.ontGraph,[EX_NS.NoExclusion])
        self.failUnless(not posRules,"There should be no rules in the 0 strata!")
        self.assertEqual(len(negRules),2,"There should be 2 'negative' rules")
        Individual.factoryGraph = network.inferredFacts
        targetClass = Class(EX_NS.NoExclusion,skipOWLClassMembership=False)
        self.failUnless(individual1 in targetClass.extent,
        "There is a BNode that bears the contains relation with another individual that is not a member of Exclusion!")
        self.assertEquals(len(list(targetClass.extent)),1,
                          "There should only be one member in NoExclusion")

class NegatedDisjunctTest(unittest.TestCase):
    def setUp(self):
        self.ontGraph = Graph()
        self.ontGraph.bind('ex', EX_NS)
        self.ontGraph.bind('owl', OWL_NS)
        Individual.factoryGraph = self.ontGraph
                
    def testStratified(self):
        bar=EX.Bar
        baz=EX.Baz
        noBarOrBaz = ~(bar|baz)
        omega = EX.Omega
        foo = omega & noBarOrBaz
        foo.identifier = EX_NS.Foo
        ruleStore,ruleGraph,network=SetupRuleStore(makeNetwork=True)
        individual=BNode()
        omega.extent = [individual]
        NormalFormReduction(self.ontGraph)
        self.assertEqual(repr(foo),
                         "ex:Omega that ( not ex:Bar ) and ( not ex:Baz )")
        posRules,negRules=CalculateStratifiedModel(network,self.ontGraph,[EX_NS.Foo])
        foo.graph = network.inferredFacts
        self.failUnless(not posRules,"There should be no rules in the 0 strata!")
        self.assertEqual(repr(negRules[0]),"Forall ?X ( ex:Foo(?X) :- And( ex:Omega(?X) not ex:Bar(?X) not ex:Baz(?X) ) )")
        self.failUnless(len(negRules)==1,"There should only be one negative rule in a higher strata")
        self.failUnless(individual in foo.extent,
                        "%s should be a member of ex:Foo"%individual)

class NegationOfAtomicConcept(unittest.TestCase):
    def setUp(self):
        self.ontGraph = Graph()
        self.ontGraph.bind('ex', EX_NS)
        self.ontGraph.bind('owl', OWL_NS)
        Individual.factoryGraph = self.ontGraph
                
    def testAtomicNegation(self):
        bar=EX.Bar
        baz=~bar
        baz.identifier = EX_NS.Baz
        ruleStore,ruleGraph,network=SetupRuleStore(makeNetwork=True)
        individual=BNode()
        individual2=BNode()
        (EX.OtherClass).extent = [individual]
        bar.extent = [individual2]
        NormalFormReduction(self.ontGraph)
        self.assertEqual(repr(baz),
                         "Class: ex:Baz DisjointWith ex:Bar\n")
        posRules,negRules=CalculateStratifiedModel(network,self.ontGraph,[EX_NS.Foo])
        self.failUnless(not posRules,"There should be no rules in the 0 strata!")
        self.failUnless(len(negRules)==1,"There should only be one negative rule in a higher strata")
        self.assertEqual(repr(negRules[0]),
                         "Forall ?X ( ex:Baz(?X) :- not ex:Bar(?X) )")        
        baz.graph = network.inferredFacts
        self.failUnless(individual in baz.extent,
                        "%s should be a member of ex:Baz"%individual)
        self.failUnless(individual2 not in baz.extent,
                        "%s should *not* be a member of ex:Baz"%individual2)
        
if __name__ == '__main__':
    unittest.main()



from FuXi.Syntax.InfixOWL import OWL_NS
from cStringIO import StringIO
from FuXi.Horn.HornRules import Clause, Ruleset, Rule, HornFromN3
from rdflib import URIRef, RDF, RDFS, Namespace, Variable, Literal, URIRef

LIST_MEMBERSHIP_SEMANTICS=\
"""
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>.
@prefix list: <http://www.w3.org/2000/10/swap/list#>.

{?L rdf:first ?I} => {?I list:in ?L}.
{?L rdf:rest ?R. ?I list:in ?R} => {?I list:in ?L}.
"""

NOMINAL_SEMANTICS=\
"""
@prefix owl: <http://www.w3.org/2002/07/owl#>.
@prefix rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>.
@prefix list: <http://www.w3.org/2000/10/swap/list#>.

#For OWL/oneOf
{?C owl:oneOf ?L. ?X list:in ?L} => {?X a ?C}.
"""

FUNCTIONAL_SEMANTCS=\
"""
@prefix owl: <http://www.w3.org/2002/07/owl#>.
@prefix log: <http://www.w3.org/2000/10/swap/log#>.

#Inverse functional semantics
{?P a owl:FunctionalProperty. ?S ?P ?O. ?S ?P ?Y. ?O log:notEqualTo ?Y } => {?O = ?Y}.
{?P a owl:InverseFunctionalProperty. ?S ?P ?O. ?Y ?P ?O. ?S log:notEqualTo ?Y } => {?S = ?Y}.

#owl:sameAs is symmetric, transitive, (and irreflexive?) and supports "smushing."
{?T1 = ?T2} => {?T2 = ?T1}.
{?T1 = ?T2. ?S = ?T1} => {?S = ?T2}.
{?T1 ?P ?O. ?T1 = ?T2.} => {?T2 ?P ?O}.
"""

DIFFERENT_FROM_SEMANTICS=\
"""
@prefix owl: <http://www.w3.org/2002/07/owl#>.
@prefix log: <http://www.w3.org/2000/10/swap/log#>.
@prefix list: <http://www.w3.org/2000/10/swap/list#>.

{ ?ANY a owl:AllDifferent; owl:distinctMembers ?L. ?L1 list:in ?L. ?L2 list:in ?L. ?L1 log:notEqualTo ?L2 } => { ?L1 owl:differentFrom ?L2 }.
"""

FUNCTIONAL_PROPERTIES=\
"""
ASK {
  [] a ?KIND
  FILTER(
      ?KIND = owl:InverseFunctionalProperty ||
      ?KIND = owl:FunctionalProperty
  )
}"""

def AdditionalRules(tBox):
    """
    Only include list and oneOf semantics
    if oneOf axiom is detected in graph 
    reduce computational complexity.  Same with other conditional axioms
    
    """
    ruleSrc = set()
    addListSemantics = False
    if tBox.query(FUNCTIONAL_PROPERTIES,
                  initNs={"owl":OWL_NS}).askAnswer[0]:
        ruleSrc.add(FUNCTIONAL_SEMANTCS)
    if (None,OWL_NS.oneOf,None) in tBox:
        ruleSrc.add(NOMINAL_SEMANTICS)
        addListSemantics = True
    if (None,RDF.type,OWL_NS.AllDifferent) in tBox:
        ruleSrc.add(DIFFERENT_FROM_SEMANTICS)
        addListSemantics = True
    if addListSemantics:
        ruleSrc.add(LIST_MEMBERSHIP_SEMANTICS)                
    for src in ruleSrc:
        for rule in HornFromN3(StringIO(src)):
            yield rule